/*
 * sonar.c
 *
 * SONAR Client Library version 0
 *
 * Jason Cox
 *
 * This file contains the complete source for the sonar client library.
 * 
 */


#include "sonar.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <errno.h>

#ifdef _AIX
#include <sys/select.h>
#endif

#define TRUE              1
#define FALSE             0
#define SONAR_REQUEST     0


static struct sockaddr_in server;
static int sockfd;
static int version;
static int request_type;
static int request_id;
static unsigned int pack_seq = 0;
static char packet[2048];
static fd_set fds_master;
static int initialized = FALSE;


/*
 * sonar_init
 *
 * Description: This function tries to find a SONAR server and sets
 *              up a udp socket to it.  This function can be called
 *              more than once to allow for changing the server.
 *              We may need to add some kind of handshake with the
 *              SONAR server to make sure the daemon is running
 *              because right now the function can't detect the
 *              server.
 *
 * Input:       NULL     - uses environment variable, rcfile, etc.
 *              a string - uses the string for sonar server host.
 *
 * Returns:     0 - success
 *             -1 - failure
 *
 */
int
sonar_init(sonarserver)
char *sonarserver;
{
    char *name, string[1024];
    struct hostent *dest;
    FILE *fp;
    struct sockaddr_in client;


    if (initialized) {
	close(sockfd);
	initialized = FALSE;
    }

    if (sonarserver) {
	name = sonarserver;
    }
    else if ((name = getenv(SONAR_ENV_NAME)) == NULL) {
	if ((name = getenv("HOME")) != NULL) {
	    if (strlen(name)+strlen(SONAR_RCFILE) < 1024) {
		strcpy(string, name);
		strcpy(string+strlen(name), SONAR_RCFILE);
	    }
	    else
		string[0] = '\0';
	}
	
	if ((fp = fopen(string, "r")) != NULL ||
	    (fp = fopen(SONAR_FILE, "r")) != NULL) {
	    if (fgets(string, 1024, fp) == NULL)
		*string = '\0';
	    else if (string[strlen(string)-1] == '\n')
		string[strlen(string)-1] = '\0';
	    name = string;
	    fclose(fp);
	}
	else
	    name = SONAR_HOST_NAME;
    }
    
    bzero((char *)&server, sizeof(server));
    if ((dest = gethostbyname(name)) == NULL &&
	(dest = gethostbyaddr(name, strlen(name), AF_INET)) == NULL)
	return SONAR_ERROR;
    
    bcopy(dest->h_addr_list[0],(char *)&server.sin_addr,dest->h_length);
    server.sin_family = AF_INET;
    server.sin_port = htons(SONAR_PORT);

    bzero((char *)&client, sizeof(client));
    client.sin_family = AF_INET;
    client.sin_port = htons(0);
    client.sin_addr.s_addr = htonl(INADDR_ANY);
    
    if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	perror("sonar_init: can't open datagram socket");
	return SONAR_ERROR;
    }

    if (bind(sockfd, (struct sockaddr *)&client, sizeof(client)) <0) {
	perror("sonar_init: can't connect to server");
	close(sockfd);
	return SONAR_ERROR;
    }

    FD_ZERO(&fds_master);
    FD_SET(sockfd, &fds_master);

    version = htonl(SONAR_VERSION);
    request_type = htonl(SONAR_REQUEST);
    
    initialized = TRUE;
    return TRUE;
}


/*
 * sonar_probe
 *
 * Description: sonar_probe takes an array of host addresses and queries
 *              a SONAR server.  This function is the only one that needs
 *              to be called by a client program to use a SONAR server.
 *              It handles initialization, so there is no need to call
 *              sonar_init except to change the SONAR server host.
 *
 * Input:   host         - array of host addresses
 *          cnt          - number of hosts in the host array
 *          timeout_sec  - Maximum number of seconds to wait
 *          timeout_usec - Maximum number of useconds to wait
 *
 * Returns: host         - array of host addresses with SONAR metrics
 *          cnt          - number of hosts
 *
 */
int
sonar_probe(host, cnt, timeout_sec, timeout_usec)
struct sonar_info *host;
int *cnt;
unsigned int timeout_sec,
             timeout_usec;
{
    fd_set fds;
    char *ptr;
    int t, i, size, try_again;
    struct timeval time, start_time;
    struct sockaddr_in recv_addr;

    
    if (*cnt < 1) {
	*cnt = 0;
	return SONAR_SUCCESS;
    }

    if (*cnt > MAX_SONAR_REQS)
	*cnt = MAX_SONAR_REQS;

    if (!initialized && sonar_init(NULL) != TRUE)
	    return SONAR_NOSERVER;
    
    /*
     * Build packet
     */
    ptr = packet;
    bcopy((char *) &version, ptr, 4);
    ptr += 4;
    
    bcopy((char *) &request_type, ptr, 4);
    ptr += 4;
    
    request_id = (getpid() & 0x0000ffff) | ((pack_seq++ & 0x0000ffff) << 16);
    bcopy((char *) &request_id, ptr, 4);
    ptr += 4;

    /* timeout */
    t = htonl(timeout_sec);
    bcopy((char *)&t, ptr, 4);
    ptr += 4;
    t = htonl(timeout_usec);
    bcopy((char *)&t, ptr, 4);
    ptr += 4;

    /* count */
    t = htonl(*cnt);
    bcopy((char *) &t, ptr, 4);
    ptr += 4;

    /* hosts */
    for (t = 0; t < *cnt; t++, ptr+=4)
	bcopy((char *) &host[t].host, ptr, 4);

    size = 24 + (*cnt * 4);
    if (sendto(sockfd, packet, size, 0, (struct sockaddr *)&server,
	       sizeof(server)) < 0) {
	perror("send");
	*cnt = 0;
	return SONAR_ERROR;
    }

    gettimeofday(&start_time, NULL);

    do {
	fds = fds_master;

	time.tv_sec = timeout_sec + 1;
	time.tv_usec = timeout_usec;
    	try_again = FALSE;

	if (select(sockfd + 1, &fds, NULL, NULL, &time) < 0) {
	    if (errno != EINTR) {
		perror("select");
		*cnt = 0;
		return SONAR_ERROR;
	    }
	}
	if (FD_ISSET(sockfd, &fds)) {
	    if ((t = recvfrom(sockfd,packet,1536,0,
			      (struct sockaddr *)&recv_addr,&size)) < 0) {
		perror("recvfrom");
		*cnt = 0;
		return SONAR_ERROR;
	    }
	    
	    if (*((int *)packet+2) != request_id) {
		try_again = TRUE;

		/* Adjust time-out */
		gettimeofday(&time, NULL);
		timeout_sec -= (time.tv_sec - start_time.tv_sec) - 1;
		start_time.tv_sec = time.tv_sec;
		start_time.tv_usec = time.tv_usec;
	    }
	    else {
		/* Check response type */
		t = ntohl(*((int *)packet+1));
		if (t != SONAR_SUCCESS) {
		    /* error response */
		    *cnt = 0;
		    return t;
		}

		ptr = packet + 12;

		/* get count */
		bcopy(ptr, &t, 4);
		t = ntohl(t);
		ptr += 4;
		
		/* copy host & metric into host array */
		for (i=0; i < t; i++) {
		    bcopy(ptr, &host[i].host, 4);
		    bcopy(ptr+4, &host[i].metric, 4);
		    host[i].metric = ntohl(host[i].metric);
		    ptr += 8;
		}
		*cnt = t;
		qsort(host, t, sizeof(struct sonar_info), (void *)sonar_cmp);
		return SONAR_SUCCESS;
	    }
	}
    } while (try_again);
    
    *cnt = 0;
    return SONAR_TIMEDOUT;
}


int
sonar_cmp(host1, host2)
struct sonar_info *host1, *host2;
{
    if (host1->metric < host2->metric)
	return -1;
    if (host1->metric > host2->metric)
	return 1;
    return 0;
}
