#include "initialize.h"

/*
 * initialize
 *
 * Description: This function performs general initialization of sockets,
 *              data structures, and options.
 */
void
initialize()
{
    extern int ident, packsize;
    extern struct sockaddr_in dest;
    extern u_char sendpack[64];
    extern int icmp_sock;
    extern int udp_sock;
    extern int sockaddr_in_size;
    extern Client clients;
    extern Hash host_data;
    extern Masklist query_mask;
    extern int *broadcast_addrs;
    
    struct icmp *icp;
    struct protoent *proto;
    struct sockaddr_in udp_server;
    int i;
    
    /* Initialize data structures */
    host_data = hash_create(HASH_TABLE_SIZE, MAX_DB_ENTRIES);
    clients = client_create(MAX_CLIENTS);
    client_update_setmem(MAX_CLIENTS * MAX_PACK_REQS);
    broadcast_addrs = findbroadcast();

    if (query_mask == NULL) {
	masktable_create();
#ifdef _DEBUG_
	fprintf(stderr, "reading from rc file\n");
#endif
	query_mask = masklist_create();
	load_config_file();
    }

    if (ml_empty(query_mask)) {
#ifdef _DEBUG_
	fprintf(stderr, "*NO MASK* using local net mask\n");
#endif
	
    }
    
    /*
     * Setup packets and sockets
     */
    sockaddr_in_size = sizeof(struct sockaddr_in);
    ident = getpid() & 0xffff;
    
    icp = (struct icmp *) sendpack;  /* pointer to ICMP header */
    icp->icmp_type  = ICMP_ECHO;
    icp->icmp_code  = 0 ;
    icp->icmp_id    = ident;
    icp->icmp_seq   = 0 ;  /* set seq # to zero for now */   

    packsize = SIZE_ICMP_HDR + sizeof(struct timeval);
    
    if ( (proto = getprotobyname("icmp")) == NULL) {
	fprintf(stderr,"unknown protocol: icmp");
	exit(errno);
    }
    
    if ( (icmp_sock = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0){
	perror("can't create raw socket, permission problem???");
	exit(errno);
    }


    /* set the icmp socket to be nonblocking */
    if ((i=fcntl(icmp_sock,F_SETFL,O_NDELAY))<0) {
	perror("unable to set icmp socket for nonblocking");
	exit(errno);
    }

    bzero((char *) &dest, sizeof(dest));
    dest.sin_family = AF_INET;
    /* udp stuff */
    if ( (udp_sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	perror("Trouble on udp socket creation");
	exit(errno);
    }

    /* set the udp socket to be nonblocking */
    if ((i=fcntl(udp_sock,F_SETFL,O_NDELAY))<0) {
	perror("unable to set udp socket for nonblocking");
	exit(1);
    }

    bzero((char *) &udp_server, sizeof(struct sockaddr));
    udp_server.sin_family = AF_INET;
    udp_server.sin_addr.s_addr = INADDR_ANY;
    udp_server.sin_port = htons(SONAR_PORT);

    if (bind(udp_sock,(struct sockaddr *)&udp_server,sizeof(udp_server)) < 0){
	perror("Trouble on udp bind");
	exit(errno);
    }
}


/*
 * load_config_file
 *
 * Description: This function reads and interprets the sonard config file
 *              if present.
 */
void
load_config_file()
{
    extern Masklist query_mask;
    extern unsigned int masktable[];
    FILE *fp;
    Addrlist addrs;
    char buf[BUFSIZ], cmd[BUFSIZ], *ptr, *ptr2;
    int len,i;
    
    if ((fp = fopen(CONFIG_FILE, "r")) != NULL) {
	while (fgets(buf, sizeof(buf), fp) != NULL) {
	    cmd[0] = '\0';
	    sscanf(buf, "%s", cmd);
	    buf[strlen(buf)-1] = '\0';
	    
	    if (!strcasecmp(cmd, "mask")) {
		ptr = buf + 5;
		if (!(ptr = strtok(ptr, "/"))) {
		    fprintf(stderr, "sonard: %s: Mask error in config file\n",
			    CONFIG_FILE);
		    exit(1);
		}
		if (!(ptr2 = strtok(NULL, "/"))) {
		    fprintf(stderr,
			    "sonard: %s: %s: Bad mask\n",CONFIG_FILE, ptr);
		    exit(1);
		}
		sscanf(ptr2, "%d", &len);
		if (len > 32 || len < 0) {
		    fprintf(stderr, "sonard: %s: %d: Invalid mask length\n",
			    CONFIG_FILE, len);
		    exit(1);
		}
		masklist_insert(query_mask, inet_addr(ptr), masktable[len]);
	    }
	}
	fclose(fp);
    }

    if (ml_empty(query_mask)) {
	/* use local networks as default query mask */
	addrs = findnets();
	for (i=0; addrs[i].addr != 0; i++)
	    masklist_insert(query_mask, addrs[i].addr, addrs[i].mask);
	masklist_insert(query_mask, 0x7f000001, 32); /* localhost */
	return;
    }
}
